from keras import layers, models
from keras.utils.generic_utils import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion
from tfreplknet.ffn import FFN
from tfreplknet.block import Block


@register_keras_serializable(package='TFRepLKNet')
class Stage(layers.Layer):
    def __init__(self, kernel_size, small_kernel, ratio, dropout, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        if not isinstance(dropout, (list, tuple)):
            raise ValueError(f'The "dropout" argument must be a list of floats. Got: {dropout}')

        self.kernel_size = kernel_size
        self.small_kernel = small_kernel
        self.ratio = ratio
        self.dropout = dropout

    @shape_type_conversion
    def build(self, input_shape):
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimension of the inputs should be defined. Found `None`.')
        self.input_spec = layers.InputSpec(ndim=4, axes={-1: channels})

        # noinspection PyAttributeOutsideInit
        self.blocks = models.Sequential()
        for i, dropout in enumerate(self.dropout):
            self.blocks.add(Block(self.kernel_size, self.small_kernel, dropout, name=f'{self.name}/blocks/{i * 2}'))
            self.blocks.add(FFN(self.ratio, dropout, name=f'{self.name}/blocks/{i * 2 + 1}'))

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        outputs = self.blocks(inputs)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({
            'kernel_size': self.kernel_size,
            'small_kernel': self.small_kernel,
            'ratio': self.ratio,
            'dropout': self.dropout
        })

        return config



from keras import layers, models
from keras.utils.generic_utils import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFRepLKNet')
class LargeConv(layers.Layer):
    def __init__(self, kernel_size, small_kernel, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.kernel_size = kernel_size
        self.small_kernel = small_kernel

    @shape_type_conversion
    def build(self, input_shape):
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimension of the inputs should be defined. Found `None`.')
        self.input_spec = layers.InputSpec(ndim=4, axes={-1: channels})

        # noinspection PyAttributeOutsideInit
        self.big = models.Sequential([
            layers.DepthwiseConv2D(
                self.kernel_size, padding='same', use_bias=False, name=f'{self.name}/lkb_origin/conv'),
            layers.BatchNormalization(momentum=0.1, epsilon=1.001e-5, name=f'{self.name}/lkb_origin/bn')
        ], name='lkb_origin')

        # noinspection PyAttributeOutsideInit
        self.small = models.Sequential([
            layers.DepthwiseConv2D(
                self.small_kernel, padding='same', use_bias=False, name=f'{self.name}/small_conv/conv'),
            layers.BatchNormalization(momentum=0.1, epsilon=1.001e-5, name=f'{self.name}/small_conv/bn')
        ], name='small_conv')

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        outputs = self.big(inputs)
        outputs += self.small(inputs)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({
            'kernel_size': self.kernel_size,
            'small_kernel': self.small_kernel
        })

        return config



from keras import activations, layers, models
from keras.utils.generic_utils import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion
from tfreplknet.large import LargeConv
from tfreplknet.drop import DropPath


@register_keras_serializable(package='TFRepLKNet')
class Block(layers.Layer):
    def __init__(self, kernel_size, small_kernel, dropout, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.kernel_size = kernel_size
        self.small_kernel = small_kernel
        self.dropout = dropout

    @shape_type_conversion
    def build(self, input_shape):
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimension of the inputs should be defined. Found `None`.')
        self.input_spec = layers.InputSpec(ndim=4, axes={-1: channels})

        # noinspection PyAttributeOutsideInit
        self.bn = layers.BatchNormalization(momentum=0.1, epsilon=1.001e-5, name='prelkb_bn')

        # noinspection PyAttributeOutsideInit
        self.pw1 = models.Sequential([
            layers.Conv2D(channels, 1, use_bias=False, name=f'{self.name}/pw1/conv'),
            layers.BatchNormalization(momentum=0.1, epsilon=1.001e-5, name=f'{self.name}/pw1/bn'),
            layers.ReLU()
        ], name='pw1')

        # noinspection PyAttributeOutsideInit
        self.large = LargeConv(self.kernel_size, self.small_kernel, name=f'{self.name}/large_kernel')

        # noinspection PyAttributeOutsideInit
        self.pw2 = models.Sequential([
            layers.Conv2D(channels, 1, use_bias=False, name=f'{self.name}/pw2/conv'),
            layers.BatchNormalization(momentum=0.1, epsilon=1.001e-5, name=f'{self.name}/pw2/bn')
        ], name='pw2')

        # noinspection PyAttributeOutsideInit
        self.drop = DropPath(self.dropout)

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        outputs = self.bn(inputs)
        outputs = self.pw1(outputs)
        outputs = self.large(outputs)
        outputs = activations.relu(outputs)
        outputs = self.pw2(outputs)
        outputs = inputs + self.drop(outputs)

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({
            'kernel_size': self.kernel_size,
            'small_kernel': self.small_kernel,
            'dropout': self.dropout
        })

        return config



from keras import layers
from keras.utils.conv_utils import normalize_tuple
from keras.utils.generic_utils import register_keras_serializable
from keras.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFRepLKNet')
class SamePad(layers.Layer):
    def __init__(self, kernel_size, dilation_rate=1, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=4)

        self.kernel_size = normalize_tuple(kernel_size, 2, 'kernel_size')
        self.dilation_rate = normalize_tuple(dilation_rate, 2, 'dilation_rate')

    @shape_type_conversion
    def build(self, input_shape):
        total_pad = (self.kernel_size[0] - 1) * self.dilation_rate[0], \
                    (self.kernel_size[1] - 1) * self.dilation_rate[1]

        top_pad = total_pad[0] // 2
        bottom_pad = total_pad[0] - top_pad
        left_pad = total_pad[1] // 2
        right_pad = total_pad[1] - top_pad

        # noinspection PyAttributeOutsideInit
        self.pad = layers.ZeroPadding2D(((top_pad, bottom_pad), (left_pad, right_pad)))

        super().build(input_shape)

    def call(self, inputs):
        return self.pad(inputs)

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return self.pad.compute_output_shape(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({
            'kernel_size': self.kernel_size,
            'dilation_rate': self.dilation_rate
        })

        return config



from tfreplknet.model import RepLKNet, RepLKNet31B, RepLKNet31L
from tfreplknet.model import RepLKNet31B224K1, RepLKNet31B224K21
from tfreplknet.model import RepLKNet31B384K1  # , RepLKNet31B384K21
from tfreplknet.model import RepLKNet31L384K1, RepLKNet31L384K21
from tfreplknet.prep import preprocess_input
